@@ -34,7 +34,7 @@ except AttributeError:
             pass
 
 __author__ = 'Alec Thomas <alec@swapoff.org>'
-__version__ = '0.7.7'
+__version__ = '0.7.8'
 __version_tag__ = ''
 
 log = logging.getLogger('injector')
@@ -44,6 +44,11 @@ if log.level == logging.NOTSET:
     log.setLevel(logging.WARN)
 
 
+def private(something):
+    something.__private__ = True
+    return something
+
+
 def synchronized(lock):
     def outside_wrapper(function):
         @functools.wraps(function)
@@ -92,21 +97,23 @@ class CallError(Error):
         if len(self.args) == 1:
             return self.args[0]
 
-        instance, method, args, kwargs, original_error = self.args
+        instance, method, args, kwargs, original_error, stack = self.args
         if hasattr(method, 'im_class'):
             instance = method.__self__
             method_name = method.__func__.__name__
         else:
             method_name = method.__name__
 
-        full_method = '.'.join((repr(instance) if instance else '', method_name)).strip('.')
+        cls = instance.__class__.__name__ if instance is not None else ''
+
+        full_method = '.'.join((cls, method_name)).strip('.')
 
         parameters = ', '.join(itertools.chain(
             (repr(arg) for arg in args),
             ('%s=%r' % (key, value) for (key, value) in kwargs.items())
         ))
-        return 'Call to %s(%s) failed: %s' % (
-            full_method, parameters, original_error)
+        return 'Call to %s(%s) failed: %s (injection stack: %r)' % (
+            full_method, parameters, original_error, [level[0] for level in stack])
 
 
 class CircularDependency(Error):
@@ -158,6 +165,7 @@ class InstanceProvider(Provider):
         return self._instance
 
 
+@private
 class ListOfProviders(Provider):
     """Provide a list of instances via other Providers."""
 
@@ -189,7 +197,7 @@ class MapBindProvider(ListOfProviders):
         return map
 
 
-# These classes are used internally by the Binder.
+@private
 class BindingKey(tuple):
     """A key mapping to a Binding."""
 
@@ -218,16 +226,15 @@ class BindingKey(tuple):
 _BindingBase = namedtuple('_BindingBase', 'interface annotation provider scope')
 
 
+@private
 class Binding(_BindingBase):
     """A binding from an (interface, annotation) to a provider in a scope."""
 
 
 class Binder(object):
-    """Bind interfaces to implementations.
-
-    :attr injector: Injector this Binder is associated with.
-    """
+    """Bind interfaces to implementations."""
 
+    @private
     def __init__(self, injector, auto_bind=True, parent=None):
         """Create a new Binder.
 
@@ -305,11 +312,8 @@ class Binder(object):
             )
 
     def create_binding(self, interface, to=None, annotation=None, scope=None):
-        if to is None:
-            to = interface
         provider = self.provider_for(interface, to)
-        if scope is None:
-            scope = getattr(to, '__scope__', NoScope)
+        scope = scope or getattr(to or interface, '__scope__', NoScope)
         if isinstance(scope, ScopeDecorator):
             scope = scope.scope
         return Binding(interface, annotation, provider, scope)
@@ -324,17 +328,24 @@ class Binder(object):
         elif issubclass(type(to), type):
             return ClassProvider(to, self.injector)
         elif isinstance(interface, ParameterizedBuilder):
-            builder = AssistedBuilderImplementation(self.injector, *interface)
+            builder = AssistedBuilderImplementation(self.injector, interface.interface, None, None)
             return CallableProvider(lambda: builder.build(**interface.kwargs))
         elif isinstance(interface, AssistedBuilder):
             builder = AssistedBuilderImplementation(self.injector, *interface)
             return InstanceProvider(builder)
-        elif isinstance(to, interface):
+        elif isinstance(interface, (tuple, type)) and isinstance(to, interface):
             return InstanceProvider(to)
-        elif issubclass(type(interface), type):
+        elif issubclass(type(interface), type) or isinstance(interface, (tuple, list)):
             if issubclass(interface, (BaseKey, BaseMappingKey, BaseSequenceKey)):
                 return InstanceProvider(to)
             return ClassProvider(interface, self.injector)
+        elif hasattr(interface, '__call__'):
+            function = to or interface
+            if hasattr(function, '__bindings__'):
+                function = self.injector.wrap_function(function)
+
+            return InstanceProvider(function)
+
         else:
             raise UnknownProvider('couldn\'t determine provider for %r to %r' %
                                   (interface, to))
@@ -511,7 +522,7 @@ class Injector(object):
         """
         # Stack of keys currently being injected. Used to detect circular
         # dependencies.
-        self._stack = []
+        self._stack = ()
 
         self.parent = parent
         self.use_annotations = use_annotations
@@ -584,8 +595,11 @@ class Injector(object):
         try:
             self.install_into(instance)
         except AttributeError:
-            # Some builtin types can not be modified.
-            pass
+            if hasattr(instance, '__slots__'):
+                raise Error('Can\'t create an instance of type %r due to presence of __slots__, '
+                            'remove __slots__ to fix that' % (cls,))
+
+            # Else do nothing - some builtin types can not be modified.
         try:
             instance.__init__(**additional_kwargs)
         except TypeError as e:
@@ -594,8 +608,7 @@ class Injector(object):
             reraise(e, CallError(
                 instance,
                 getattr(instance.__init__, '__func__', instance.__init__),
-                (), additional_kwargs, e,
-                )
+                (), additional_kwargs, e, self._stack,)
             )
         return instance
 
@@ -611,7 +624,27 @@ class Injector(object):
         """
         instance.__injector__ = self
 
-    def call_with_injection(self, callable, self_, args=(), kwargs={}):
+    @private
+    def wrap_function(self, function):
+        """Create function wrapper that will take care of it's dependencies.
+
+        You only need to provide noninjectable arguments to the wrapped function.
+
+        :return: Wrapped function.
+        """
+
+        @functools.wraps(function)
+        def wrapper(*args, **kwargs):
+            return self.call_with_injection(
+                callable=function,
+                args=args,
+                kwargs=kwargs
+            )
+
+        return wrapper
+
+    @private
+    def call_with_injection(self, callable, self_=None, args=(), kwargs={}):
         """Call a callable and provide it's dependencies if needed.
 
         :param self_: Instance of a class callable belongs to if it's a method,
@@ -639,8 +672,9 @@ class Injector(object):
                 *((self_,) if self_ is not None else ()) + tuple(args),
                 **dependencies)
         except TypeError as e:
-            reraise(e, CallError(self_, callable, args, dependencies, e))
+            reraise(e, CallError(self_, callable, args, dependencies, e, self._stack))
 
+    @private
     @synchronized(lock)
     def args_to_inject(self, function, bindings, owner_key):
         """Inject arguments into a function.
@@ -666,7 +700,7 @@ class Injector(object):
                 repr_key(key)),
                 )
 
-        self._stack.append(key)
+        self._stack += (key,)
         try:
             for arg, key in bindings.items():
                 try:
@@ -680,7 +714,7 @@ class Injector(object):
                     raise e
                 dependencies[arg] = instance
         finally:
-            self._stack.pop()
+            self._stack = tuple(self._stack[:-1])
 
         return dependencies
 
@@ -814,8 +848,15 @@ def inject(**bindings):
 
         @inject(**bindings)
         def init(self, *args, **kwargs):
-            for key in original_keys:
-                setattr(self, key, kwargs.pop(key.lstrip('_')))
+            try:
+                for key in original_keys:
+                    normalized_key = key.lstrip('_')
+                    setattr(self, key, kwargs.pop(normalized_key))
+            except KeyError as e:
+                reraise(e, CallError(
+                    'Keyword argument %s not found when calling %s' % (
+                        normalized_key, '%s.%s' % (cls.__name__, '__init__'))))
+
             orig_init(self, *args, **kwargs)
         cls.__init__ = init
         return cls
